/* $Id: kanji.c,v 1.29 2005/01/10 03:49:32 dtana Exp $ */
/*
 *		Kanji handling routines.
 *		These are only used when kanji is #defined.
 *
 *		Coded by Shigeki Yoshida (shige@csk.CO.JP)
 */
#include	"config.h"	/* 90.12.20  by S.Yoshida */
#ifdef HAVE_ICONV
#include <iconv.h>
#endif
#include	"def.h"
#include	"kinit.h"

int	global_kexpect	= KEXPECT;	/* kanji-expected-code		*/
int	global_kinput 	= KINPUT;	/* kanji-input-code		*/
static int display_encoding = KDISPLAY;
static int fileinp_encoding = KFIO;
static int local_kfin;			/* Buffer local file input code. */

/*
 * Input one byte from the keyboard with KANJI code conversion.
 */
enum {
	SELROMA  = 0,
	SELKANJI,
	SELKANA,
	SELHOJO
};

static struct charstate {
	int charset, kana;
	int to_kanji, to_ascii, to_kana;
	unsigned char b[4];
} display_state = {
	SELROMA, FALSE,
	TO_KDISPLAY, TO_ADISPLAY, TO_KANADISPLAY,
	{0},
}, fileio_state = {
	SELROMA, FALSE,
	TO_KFIO, TO_AFIO, TO_KANAFIO,
	{0},
};

static char symbol_c[] = {'N', 'S', 'J', 'E', 'U', '-', 'T'};
static char *symbol_s[] = {
	"No-conversion", "Shift-JIS", "JIS", "EUC", "UTF", "NIL", "T"};
/* Symbol chars & strings to	*/
/* display a kanji code info to	*/
/* the user.  This order depend	*/
/* on the values of kanji code	*/
/* macro defined at def.h.	*/
static char *kcodename_u[] = {
	"NOCONV", "SHIFT-JIS", "JIS", "EUC", "UTF", "NIL", "T"};
static char *kcodename_l[] = {
	"noconv", "shift-jis", "jis", "euc", "utf", "nil", "t"};
#define	NKCODENAME	7
					/* The strings of kanji code	*/
					/* name and the numner of kanji	*/
					/* code to determine what kanji	*/
					/* code the user want to use.	*/
					/* This order depend on	the	*/
					/* values of kanji code macro	*/
					/* defined at def.h.		*/

#define	ESC	CCHR('[')		/* Escape char.			*/

#define	issjis1st(c)	(((c) >= 0x81 && (c) <= 0x9f) || \
			 ((c) >= 0xe0 && (c) <= 0xfc))
#define issjis2nd(c) ((0x40<=(c)&&(c)<=0x7e)||(0x80<=(c)&&(c)<=0xfc))
#define	iseuc1st(c)	((c) >= 0xa1 && (c) <= 0xfe)
#define iseuc2nd(c) iseuc1st(c)
#define iskana(c)	((c) >= 0xa0 && (c) < 0xe0)


/* 90.07.25  Change to inline routine  by S.Yoshida (from) */
#define	jtoe(c1, c2)	{c1 |= 0x80; c2 |= 0x80;}
#define	etoj(c1, c2)	{c1 &= 0x7f; c2 &= 0x7f;}
#define	stoe(c1, c2)	\
{\
	if (c1 >=0xe0)\
		c1 -=  0x40;\
	if (c2 >= 0x9f) {\
		c1 = (c1 - 0x88) * 2 + 0xb0;\
		c2 += 0x02;\
	} else {\
		if (c2 >= 0x7f)\
			c2 -= 0x01;\
		c1 = (c1 - 0x89) * 2 + 0xb1;\
		c2 = c2 + 0x61;\
	}\
}
#define	etos(c1, c2)	\
{\
	c1 &= 0x7f;\
	c2 &= 0x7f;\
	if(c1 >= 0x5f)\
		c1 += 0x80;\
	if((c1 % 2) == 0) {\
		c1 = (c1 - 0x30) / 2 + 0x88;\
		c2 += 0x7e;\
	} else {\
		if(c2 >= 0x60)\
			c2 += 0x01;\
		c1 = (c1 - 0x31) / 2 + 0x89;\
		c2 += 0x1f;\
	}\
	c1 &= 0xff;\
	c2 &= 0xff;\
}
/* 90.07.25  Change to inline routine  by S.Yoshida (to) */

#define	CAT_NONE	0		/* Category: No category.	*/
#define	CAT_ASCII	1		/* Category: ASCII code char.	*/
#define	CAT_KIGOU	2		/* Category: JIS KIGOU char.	*/
#define	CAT_EISUUJI	3		/* Category: JIS EI-SUUJI char.	*/
#define	CAT_HIRAGANA	4		/* Category: HIRAGANA char.	*/
#define	CAT_KATAKANA	5		/* Category: KATAKANA char.	*/
#define	CAT_GREEK	6		/* Category: JIS GREEK char.	*/
#define	CAT_RUSSIAN	7		/* Category: JIS RUSSIAN char.	*/
#define	CAT_KANJI	8		/* Category: kanji char.	*/

static int
etou(char *u, const char *e)
{
#ifdef HAVE_ICONV
	char *obuff;
	size_t isize, osize;
	static iconv_t cd = (iconv_t)-1;

	if (cd == (iconv_t)-1) {
		if ((cd = iconv_open("UTF-8", "EUC-JP")) == (iconv_t)-1)
			perror("iconv_open"), exit(1);
	}
	isize = strlen(e) + 1;
	osize = 7;
	obuff = u;
	if (iconv(cd, &e, &isize, &obuff, &osize) == (size_t)-1)
		return -1;
#else
	strcpy(u, e);
#endif
	return 0;
}

static int
utoe(char *u, const char *e)
{
#ifdef HAVE_ICONV
	char *obuff;
	size_t isize, osize;
	static iconv_t cd = (iconv_t)-1;

	if (cd == (iconv_t)-1) {
		if ((cd = iconv_open("EUC-JP", "UTF-8")) == (iconv_t)-1)
			perror("iconv_open"), exit(1);
	}
	isize = strlen(e) + 1;
	osize = 7;
	obuff = u;
	if (iconv(cd, &e, &isize, &obuff, &osize) == (size_t)-1)
		return -1;
#else
	strcpy(u, e);
#endif
	return 0;
}

static int
eton(char *u, const char *e)
{
#ifdef HAVE_ICONV
	char *obuff;
	size_t isize, osize;
	static iconv_t cd = (iconv_t)-1;

	if (cd == (iconv_t)-1) {
		if ((cd = iconv_open("", "EUC-JP")) == (iconv_t)-1)
			perror("iconv_open"), exit(1);
	}
	isize = strlen(e) + 1;
	osize = 7;
	obuff = u;
	if (iconv(cd, &e, &isize, &obuff, &osize) == (size_t)-1)
		return -1;
#else
	strcpy(u, e);
#endif
	return 0;
}

static int
ntoe(char *u, const char *e)
{
#ifdef HAVE_ICONV
	char *obuff;
	size_t isize, osize;
	static iconv_t cd = (iconv_t)-1;

	if (cd == (iconv_t)-1) {
		if ((cd = iconv_open("EUC-JP", "")) == (iconv_t)-1)
			perror("iconv_open"), exit(1);
	}
	isize = strlen(e) + 1;
	osize = 7;
	obuff = u;
	if (iconv(cd, &e, &isize, &obuff, &osize) == (size_t)-1)
		return -1;
#else
	strcpy(u, e);
#endif
	return 0;
}

static void
ttputs(const char *s)
{
	for (; *s; s++)
		ttputc((unsigned char)*s);
}

/* prototype for static routines */
static int kcodenumber pro((int *code, int num));
static int kanalastchar pro((int *lastch));
static int klastchar pro((int *lastch, int kselect));
static int kcodecheck pro((char *p, int len));
static int bufjtoe_c pro((char *j, int len));
int bufstoe_c pro((char *p, int len));

/*
 * COMMAND: change-default-fileio-code
 * This is a command processor for changing the value of the global
 * file I/O kanji code.  When this function is called, the value is
 * changed rotaly. (NIL -> NOCONV -> SJIS -> JIS -> EUC)
 */
/*ARGSUSED*/
k_rot_fio(f, n)
{
	if (fileinp_encoding == NIL) {
		fileinp_encoding = NOCONV;
	} else if (fileinp_encoding == NOCONV) {
		fileinp_encoding = SJIS;
	} else if (fileinp_encoding == SJIS) {
		fileinp_encoding = JIS;
	} else if (fileinp_encoding == JIS) {
		fileinp_encoding = EUC;
	} else if (fileinp_encoding == EUC) {
		fileinp_encoding = UTF;
	} else if (fileinp_encoding == UTF) {
		fileinp_encoding = NIL;
	}
	return TRUE;
}

/*
 * COMMAND: set-default-fileio-code
 * This is a command processor for setting the value of the global
 * file I/O kanji code.
 */
/*ARGSUSED*/
k_set_fio(f, n)
{
	register int	s;

	if (f & FFARG) {
		if (n >= 0 && n < NKCODENAME - 1) {
			fileinp_encoding = n;
			s = TRUE;
		} else {
			s = FALSE;
		}
	} else if ((s = kcodenumber(&n, NKCODENAME - 1)) == TRUE) {
		fileinp_encoding = n;
	}
	return (s);
}

/*
 * COMMAND: change-fileio-code
 * This is a command processor for changing the value of the buffer
 * local file I/O kanji code.  When this function is called, the value
 * is changed rotaly. (NIL -> NOCONV -> SJIS -> JIS -> EUC)
 */
/*ARGSUSED*/
k_rot_buffio(f, n)
{
	if (curbp->b_kfio == NIL) {
		curbp->b_kfio = NOCONV;
	} else if (curbp->b_kfio == NOCONV) {
		curbp->b_kfio = SJIS;
	} else if (curbp->b_kfio == SJIS) {
		curbp->b_kfio = JIS;
	} else if (curbp->b_kfio == JIS) {
		curbp->b_kfio = EUC;
	} else if (curbp->b_kfio == EUC) {
		curbp->b_kfio = UTF;
	} else if (curbp->b_kfio == UTF) {
		curbp->b_kfio = NIL;
	}
	upmodes(curbp);			/* Only update cur-buf's modeline. */
	return TRUE;
}

/*
 * COMMAND: set-kanji-fileio-code
 * This is a command processor for setting the value of the buffer
 * local file I/O kanji code.
 */
/*ARGSUSED*/
k_set_buffio(f, n)
{
	register int	s;

	if (f & FFARG) {
		if (n > 0 && n < NKCODENAME - 1) {
			curbp->b_kfio = n;
			upmodes(curbp);	/* Only update cur-buf's modeline. */
			s = TRUE;
		} else {
			s = FALSE;
		}
	} else if ((s = kcodenumber(&n, NKCODENAME - 1)) == TRUE) {
		curbp->b_kfio = n;
		upmodes(curbp);		/* Only update cur-buf's modeline. */
	}
	return (s);
}

/*
 * COMMAND: set-kanji-expected-code
 * This is a command processor for setting the value of the global
 * expected file input kanji code.
 */
/*ARGSUSED*/
k_set_expect(f, n)
{
	register int	s;

	if (f & FFARG) {
		if (n > 0 && n < NKCODENAME) {
			global_kexpect = n;
			s = TRUE;
		} else {
			s = FALSE;
		}
	} else if ((s = kcodenumber(&n, NKCODENAME)) == TRUE) {
		global_kexpect = n;
	}
	return (s);
}

/*
 * COMMAND: change-input-code
 * This is a command processor for changing the value of the global
 * keyboard input kanji code.  When this function is called, the value
 * is changed rotaly. (NOCONV -> SJIS -> JIS -> EUC)
 */
/*ARGSUSED*/
k_rot_input(f, n)
{
	if (global_kinput == NOCONV) {
		global_kinput = SJIS;
	} else if (global_kinput == SJIS) {
		global_kinput = JIS;
	} else if (global_kinput == JIS) {
		global_kinput = EUC;
	} else if (global_kinput == EUC) {
		global_kinput = UTF;
	} else if (global_kinput == UTF) {
		global_kinput = NIL;
	} else if (global_kinput == NIL) {
		global_kinput = NOCONV;
	}
	upmodes(NULL);			/* Update each win's modeline. */
	return TRUE;
}

/*
 * COMMAND: set-kanji-input-code
 * This is a command processor for setting the value of the global
 * keyboard input kanji code.
 */
/*ARGSUSED*/
k_set_input(f, n)
{
	register int	s;

	if (f & FFARG) {
		if (n > 0 && n < NKCODENAME - 1) {
			global_kinput = n;
			upmodes(NULL);	/* Update each win's modeline. */
			s = TRUE;
		} else {
			s = FALSE;
		}
	} else if ((s = kcodenumber(&n, NKCODENAME - 1)) == TRUE) {
		global_kinput = n;
		upmodes(NULL);		/* Update each win's modeline. */
	}
	return (s);
}

/*
 * COMMAND: change-display-code
 * This is a command processor for changing the value of the global
 * display kanji code.  When this function is called, the value is
 * changed rotaly. (NOCONV -> SJIS -> JIS -> EUC)
 */
/*ARGSUSED*/
k_rot_display(f, n)
{
	if (display_encoding == NOCONV) {
		display_encoding = SJIS;
	} else if (display_encoding == SJIS) {
		display_encoding = JIS;
	} else if (display_encoding == JIS) {
		display_encoding = EUC;
	} else if (display_encoding == EUC) {
		display_encoding = UTF;
	} else if (display_encoding == UTF) {
		display_encoding = NIL;
	} else if (display_encoding == NIL) {
		display_encoding = NOCONV;
	}
	sgarbf = TRUE;			/* Must update full screen. */
	return TRUE;
}

/*
 * COMMAND: set-kanji-display-code
 * This is a command processor for setting the value of the global
 * display kanji code.
 */
/*ARGSUSED*/
k_set_display(f, n)
{
	register int	s;

	if (f & FFARG) {
		if (n > 0 && n < NKCODENAME - 1) {
			display_encoding = n;
			sgarbf = TRUE;	/* Must update full screen. */
			s = TRUE;
		} else {
			s = FALSE;
		}
	} else if ((s = kcodenumber(&n, NKCODENAME - 1)) == TRUE) {
		display_encoding = n;
		sgarbf = TRUE;		/* Must update full screen. */
	}
	return (s);
}

/*
 * COMMAND: list-kanji-code
 * Display a list of kanji-code related valiables in the *Kanji Codes*
 * buffer.
 */
/*ARGSUSED*/
k_list_code(f, n)
{
	register BUFFER	*bp;
	register WINDOW	*wp;
	char	 line[80];

	if ((bp = bfind("*Kanji Codes*", TRUE)) == NULL) return FALSE;
#ifdef	AUTOSAVE	/* 96.12.24 by M.Suzuki	*/
	bp->b_flag &= ~(BFCHG | BFACHG);	/* Blow away old.	*/
#else
	bp->b_flag &= ~BFCHG;			/* Blow away old.	*/
#endif	/* AUTOSAVE	*/
	if (bclear(bp) != TRUE) return FALSE;

	strcpy(line, "* List of kanji-code related variables *");
	if (addline(bp, line) == FALSE) return FALSE;
	strcpy(line, "Global variables:");
	if (addline(bp, line) == FALSE) return FALSE;
	sprintf(line, "\tdefault-kanji-fileio-code : %s",
		symbol_s[fileinp_encoding]);
	if (addline(bp, line) == FALSE) return FALSE;
	sprintf(line, "\tkanji-expected-code : %s",
		symbol_s[global_kexpect]);
	if (addline(bp, line) == FALSE) return FALSE;
	sprintf(line, "\tkanji-display-code : %s",
		symbol_s[display_encoding]);
	if (addline(bp, line) == FALSE) return FALSE;
	sprintf(line, "\tkanji-input-code : %s",
		symbol_s[global_kinput]);
	if (addline(bp, line) == FALSE) return FALSE;
	if (curbp->b_kfio == JIS ||
	    fileinp_encoding == JIS || display_encoding == JIS) {
		sprintf(line, "\tto-kanji-fileio : %d (%c)",
			fileio_state.to_kanji, fileio_state.to_kanji);
		if (addline(bp, line) == FALSE) return FALSE;
		sprintf(line, "\tto-ascii-fileio : %d (%c)",
			fileio_state.to_ascii, fileio_state.to_ascii);
		if (addline(bp, line) == FALSE) return FALSE;
		sprintf(line, "\tto-kana-fileio : %d (%c)",
			fileio_state.to_kana, fileio_state.to_kana);
		if (addline(bp, line) == FALSE) return FALSE;
		sprintf(line, "\tto-kanji-display : %d (%c)",
			display_state.to_kanji, display_state.to_kanji);
		if (addline(bp, line) == FALSE) return FALSE;
		sprintf(line, "\tto-ascii-display : %d (%c)",
			display_state.to_ascii, display_state.to_ascii);
		if (addline(bp, line) == FALSE) return FALSE;
		sprintf(line, "\tto-kana-display : %d (%c)",
			display_state.to_kana, display_state.to_kana);
		if (addline(bp, line) == FALSE) return FALSE;
	}
	strcpy(line, "Buffer local variable:");
	if (addline(bp, line) == FALSE) return FALSE;
	sprintf(line, "\tkanji-fileio-code : %s",
		symbol_s[curbp->b_kfio]);
	if (addline(bp, line) == FALSE) return FALSE;

	if ((wp = popbuf(bp)) == NULL)	return FALSE;
	bp->b_dotp = lforw(bp->b_linep); /* put dot at beginning of buffer */
	bp->b_doto = 0;
	wp->w_dotp = bp->b_dotp;	/* fix up if window already on screen */
	wp->w_doto = bp->b_doto;
	return TRUE;
}

/*
 * COMMAND: list-kanji-code-briefly
 * Display a list of values of kanji-code related valiables in the
 * mini buffer.
 */
/*ARGSUSED*/
k_show_code(f, n)
{
	char	buf[80];
	char	*bp;

	sprintf(buf, "global:[FID=%c%c%c] local:[F=%c] expected:[E=%c]",
		symbol_c[fileinp_encoding],
		symbol_c[global_kinput],
		symbol_c[display_encoding],
		symbol_c[curbp->b_kfio],
		symbol_c[global_kexpect]);
	if (curbp->b_kfio == JIS ||
	    fileinp_encoding == JIS || display_encoding == JIS) {
		bp = buf + strlen(buf);
		sprintf(bp,
			" K:[FD=%c%c] A:[FD=%c%c] KANA:[FD=%c%c]",
			fileio_state.to_kanji, display_state.to_kanji, fileio_state.to_ascii, display_state.to_ascii,
			fileio_state.to_kana, display_state.to_kana);
	}

	ewprintf(buf);
	return TRUE;
}

/*
 * COMMAND: set-to-kanji-fileio
 * This is a command processor for setting the value of the global
 * kanji select char for a file.
 */
/*ARGSUSED*/
k_set_tokfio(f, n)
{
	int	s;

	if (f & FFARG) {
		if (n == '@' || n == 'B') {
			fileio_state.to_kanji = n;
			s = TRUE;
		} else {
			s = FALSE;
		}
	} else if ((s = klastchar(&n, TRUE)) == TRUE) {
		fileio_state.to_kanji = n;
	}
	return (s);
}

/*
 * COMMAND: set-to-ascii-fileio
 * This is a command processor for setting the value of the global
 * ASCII select char for a file.
 */
/*ARGSUSED*/
k_set_toafio(f, n)
{
	register int	s;

	if (f & FFARG) {
		if (n == 'B' || n == 'J' || n == 'H') {
			fileio_state.to_ascii = n;
			s = TRUE;
		} else {
			s = FALSE;
		}
	} else if ((s = klastchar(&n, FALSE)) == TRUE) {
		fileio_state.to_ascii = n;
	}
	return (s);
}

/*
 * COMMAND: set-to-kana-fileio
 * This is a command processor for setting the value of the global
 * KANA select char for a file.
 */
/*ARGSUSED*/
k_set_tokanafio(f, n)
{
	register int	s;

	if (f & FFARG) {
		if (n == '7' || n == '8' || n == 'I') {
			fileio_state.to_kana = n;
			s = TRUE;
		} else {
			s = FALSE;
		}
	} else if ((s = kanalastchar(&n)) == TRUE) {
		fileio_state.to_kana = n;
	}
	return (s);
}

/*
 * COMMAND: set-to-kana-display
 * This is a command processor for setting the value of the global
 * KANA select char for a display.
 */
/*ARGSUSED*/
k_set_tokanadisplay(f, n)
{
	int	s;

	if (f & FFARG) {
		if (n == '7' || n == '8' || n == 'I') {
			fileio_state.to_kana = n;
			s = TRUE;
		} else {
			s = FALSE;
		}
	} else if ((s = kanalastchar(&n)) == TRUE) {
		display_state.to_kana = n;
	}
	return (s);
}

/*
 * COMMAND: set-to-kanji-display
 * This is a command processor for setting the value of the global
 * kanji select char for a display.
 */
/*ARGSUSED*/
k_set_tokdisplay(f, n)
{
	int	s;

	if (f & FFARG) {
		if (n == '@' || n == 'B') {
			fileio_state.to_kanji = n;
			s = TRUE;
		} else {
			s = FALSE;
		}
	} else if ((s = klastchar(&n, TRUE)) == TRUE) {
		display_state.to_kanji = n;
	}
	return (s);
}

/*
 * COMMAND: set-to-ascii-display
 * This is a command processor for setting the value of the global
 * ASCII select char for a display.
 */
/*ARGSUSED*/
k_set_toadisplay(f, n)
{
	int	s;

	if (f & FFARG) {
		if (n == 'B' || n == 'J' || n == 'H') {
			fileio_state.to_ascii = n;
			s = TRUE;
		} else {
			s = FALSE;
		}
	} else if ((s = klastchar(&n, FALSE)) == TRUE) {
		display_state.to_ascii = n;
	}
	return (s);
}

/*
 * Select target kanji code and return its inner number.
 */
static int
kcodenumber(code, num)
int	*code;
int	num;
{
	register int	s;
	register int	i;
	register int	n;
	char	kcode[NINPUT];

	for (n = 0; n < 2; n++) {
		if ((s = ereply("Kanji Code System : ",
			    kcode, sizeof(kcode))) != TRUE) {
			return (s);
		}
		if (ISDIGIT(kcode[0])) {
			if ((i = atoi(kcode)) >= 0 && i < num) {
				*code = i;
				return TRUE;
			}
		} else {
			for (i = 0; i < num; i++) {
				if (strcmp(kcodename_u[i], kcode) == 0 ||
				    strcmp(kcodename_l[i], kcode) == 0) {
					*code = i;
					return TRUE;
				}
			}
		}
	}
	return FALSE;
}

/*
 * Input last char of the KANA code select escape sequence,
 * and return its value.
 */
static int
kanalastchar(lastch)
int	*lastch;
{
	register int	s;
	register int	i;
	register char	*p = "KANA Select Char [78I] : ";
	register int	n;
	char	lchar[NINPUT];

	for (n = 0; n < 2; n++) {
		if ((s = ereply(p, lchar, sizeof(lchar))) != TRUE) {
			return (s);
		}
		if (ISDIGIT(lchar[0])) {
			i = atoi(lchar);
			if (i==7) i='7';
			if (i==8) i='8';
		} else {
			i = lchar[0];
		}
		if (i == '7' || i == '8' || i == 'I') {
			*lastch = i;
			return TRUE;
		}
	}
	return FALSE;
}

/*
 * Input last char of the Kanji/ASCII code select escape sequence,
 * and return its value.  kanji select char is input when kselect
 * is TRUE, othewise ASCII select char is input.
 */
static int
klastchar(lastch, kselect)
int	*lastch;
int	kselect;
{
	register int	s;
	register int	i;
	register char	*p = kselect ? "KANJI Select Char [@B] : " :
		                       "ASCII Select Char [BHJ] : ";
	register int	n;
	char	lchar[NINPUT];

	for (n = 0; n < 2; n++) {
		if ((s = ereply(p, lchar, sizeof(lchar))) != TRUE) {
			return (s);
		}
		if (ISDIGIT(lchar[0])) {
			i = atoi(lchar);
		} else {
			i = lchar[0];
		}
		if ((kselect && (i == '@' || i == 'B')) ||
		    (!kselect && (i == 'B' || i == 'J' || i == 'H'))) {
			*lastch = i;
			return TRUE;
		}
	}
	return FALSE;
}

/*
 * Set the buffer local file I/O code when buffer is created.
 */
VOID
ksetbufcode(bp)
BUFFER	*bp;
{
	bp->b_kfio = fileinp_encoding;
}

/*
 * Show file I/O, input, display kanji code into mode line.
 */
kdispbufcode(bp)
register BUFFER	*bp;
{
	register int	n = 0;

	/* 90.12.28  Change display style like as Nemacs 3.3. by S.Yoshida */
	vtputc(symbol_c[bp->b_kfio]);		n++;
	vtputc(symbol_c[global_kinput]);	n++;
	vtputc(symbol_c[display_encoding]);	n++;
	vtputc(':');				n++;
	return (n);
}

static int
inpc_sjis(int d[])
{
	int n;

	d[0] = getkbd(), n = 1;

	if (iskana(d[0] & 0xff)) {
		d[1] = d[0], n = 2;
		d[0] = SS2;
	} else if (issjis1st(d[0])) {
		d[1] = getkbd(), n = 2;
		stoe(d[0], d[1]);
	}
	return n;
}

static int
inpc_jis(int d[])
{
	static int kselected = SELROMA;
	int n;

	d[0] = getkbd(), n = 1;
	/*
	 * If you enable next codes, you cannot input C-n,C-o in JIS
	 * input mode.  I think this state you don't want to, I
	 * disable these.  27 mar 2000, amura
	 */
	if (d[0] == ESC) {
		int c1 = getkbd();
		if (c1 == '$') {
			int c2 = getkbd();
			if (c2 == '@' || c2 == 'B') {
				kselected = SELKANJI;
				return inpc_jis(d);
			} else if (c2 == '(') {
				if ((c1 = getkbd()) == 'D') {
					kselected = SELHOJO;
					return inpc_jis(d);
				}
				else {
					ungetkbd(c1);
					ungetkbd(c2);
					ungetkbd('$');
				}
			} else {
				ungetkbd(c2);
				ungetkbd(c1);
			}
		} else if (c1 == '(') {
			int c2 = getkbd();
			if (c2 == 'B' || c2 == 'J' || c2 == 'H') {
				kselected = SELROMA;
				/* When typeahead() is TRUE, update()	*/
				/* isn't done.  So, sometimes a input	*/
				/* strings isn't displayed with JIS	*/
				/* code input.  To display the input	*/
				/* strings certainly, do update() when	*/
				/* ASCII select sequence has come.	*/
				update();
				return inpc_jis(d);

			} else if (c2 == 'I') {
				kselected = SELKANA;
				return inpc_jis(d);

			} else {
				ungetkbd(c2);
				ungetkbd(c1);
			}
		} else {
			ungetkbd(c1);
		}
	} else if (d[0] == 0xff ) {
		d[0] = 0x7f;
	} else if (kselected == SELKANJI) {
		d[0] &= 0x7f;
		if (!ISCTRL(d[0])) {
			d[1] = getkbd(), n = 2;
			jtoe(d[0], d[1]);
		}
	} else if (kselected == SELHOJO) {
		d[1] = d[0] & 0x7f;
		d[2] = getkbd(), n = 3;
		jtoe(d[1], d[2]);
		d[0] = SS3;
	} else if (kselected == SELKANA) {
		d[1] = d[0] | 0x80, n = 2;
		d[0] = SS2;
	} else if (iskana(d[0] & 0xff)) {
		d[1] = d[0], n = 2;
		d[0] = SS2;
	}
	return n;
}

static int
inpc_euc(int d[])
{
	int n;

	d[0] = getkbd(), n = 1;

	if (d[0] == SS3) {
		d[1] = getkbd();
		d[2] = getkbd();
		n = 3;
	} else if (iseuc1st(d[0]) || d[0] == SS2) {
		d[1] = getkbd(), n = 2;
	}
	return n;
}

static int utlen(int);

static int
inpc_utf(int d[])
{
	int n;

	d[0] = getkbd(), n = 1;
	if (0x80 < d[0]) {
		int length, k;
		char dm[10], es[10];

		length = utlen(d[0]);
		dm[0] = (unsigned char)d[0];
		for (k = 1; k < length; k++)
			dm[k] = (unsigned char)getkbd();
		dm[k] = 0;
		if (utoe(es, dm) < 0)
			strcpy(es, ".");
		for (k = 0; es[k]; k++)
			d[k] = (unsigned char)es[k];
		n = k;
	}
	return n;
}

static int
inpc_nil(int d[])
{
	int n;
	
	d[0] = getkbd(), n = 1;
	if (0x80 < d[0]) {
		int length, k;
		char dm[10];

		memset(dm, 0xA1, 10), dm[9] = 0;
		dm[0] = (unsigned char)d[0];
		if (0 < (length = mblen(dm, MB_CUR_MAX))) {
			char es[10];

			for (k = 1; k < length; k++)
				dm[k] = (unsigned char)getkbd();
			dm[k] = 0;
			if (ntoe(es, dm) < 0)
				strcpy(es, ".");
			for (k = 0; es[k]; k++)
				d[k] = (unsigned char)es[k];
			n = k;
		}
	}
	return n;
}

static struct fifo {
	int n, k;
	int data[10];
} inputb;

VOID
kgetkeyflush()
{
	inputb.k = inputb.n = 0;
}

int
kgetkey()
{
	if (inputb.n == inputb.k) {
		if (global_kinput == SJIS)
			inputb.n = inpc_sjis(inputb.data);
		else if (global_kinput == JIS)
			inputb.n = inpc_jis(inputb.data);
		else if (global_kinput == EUC)
			inputb.n = inpc_euc(inputb.data);
		else if (global_kinput == UTF)
			inputb.n = inpc_utf(inputb.data);
		else
			inputb.n = inpc_nil(inputb.data);
#ifdef VTCURSOR
		if (inputb.data[0] == ESC && inputb.n == 1) {
			int c1 = getkbd();
			if (c1 == 'O' || c1 == '[') {
				int c2 = getkbd(), cc;
				switch (c2) {
				case 'A':
					cc = CCHR('P');
					break;
				case 'B':
					cc = CCHR('N');
					break;
				case 'C':
					cc = CCHR('F');
					break;
				case 'D':
					cc = CCHR('B');
					break;
				default:
					ungetkbd(c2);
					ungetkbd(c1);
					cc = ESC;
					break;
				}
				inputb.data[0] = cc;
			} else 
				ungetkbd(c1);
		}
#endif
		inputb.k = 0;
	}
	return inputb.data[inputb.k++];
}

static int
rputc(int c, FILE *fp)
{
	if (fp == NULL) {
		ttputc(c);
		return 0;
	}
	return fputc(c, fp);
}

static int
rputs(const char *s, FILE *fp)
{
	for (; *s; s++)
		if (rputc((unsigned char)*s, fp) < 0)
			return -1;
	return 0;
}

static int
select_charset(struct charstate *sp, int charset, FILE *fp)
{
	if (charset == SELHOJO) {
		if (sp->kana && sp->to_kana == '7')
			if (rputc(0x0f, fp) < 0)
				return -1;
		sp->kana = FALSE;
		if (sp->charset != SELHOJO) {
			if (rputc(ESC, fp) < 0 || 
			    rputc('$', fp) < 0 || 
			    rputc('(', fp) < 0 ||
			    rputc('D', fp) < 0)
				return -1;
			sp->charset = SELHOJO;
		}
	} else if (charset == SELKANA) {
		if (!sp->kana) {
			sp->kana = TRUE;
			switch (sp->to_kana) {
			case '7':
				if (rputc(0x0e, fp) < 0)
					return -1;
				break;
			case '8':
				break;
			case 'I':
				if (rputc(ESC, fp) < 0 ||
				    rputc('(', fp) < 0 ||
				    rputc(sp->to_kana, fp) < 0)
					return -1;
				sp->charset = SELKANA;
				break;
			}
		}
	} else if (charset == SELKANJI) {
		if (sp->kana && sp->to_kana == '7')
			if (rputc(0x0f, fp) < 0)
				return -1;
		sp->kana = FALSE;
		if (sp->charset != SELKANJI) {
			if (rputc(ESC, fp) < 0 || 
			    rputc('$', fp) < 0 ||
			    rputc(sp->to_kanji, fp) < 0)
				return -1;
			sp->charset = SELKANJI;
		}
	} else {
		if (sp->kana) {
			sp->kana = FALSE;
			if (sp->to_kana == '7')
				if (rputc(0x0f, fp) < 0)
					return -1;
		}
		if (sp->charset != SELROMA) {
			if (rputc(ESC, fp) < 0 ||
			    rputc('(', fp) < 0 ||
			    rputc(sp->to_ascii, fp) < 0)
				return -1;
			sp->charset = SELROMA;
		}
	}
	return 0;
}	

static int
mbputc(struct charstate *sp, int encoding, int c, FILE *fp)
{
	char ob[7];
	int n;

	/* ASCII */
	if ((c & 0x80) == 0) {
		if (encoding == JIS)
			if (select_charset(sp, SELROMA, fp) < 0)
				return -1;
		if (rputc(c, fp) < 0)
			return -1;
		sp->b[0] = 0;
		return 1;
	}

	n = strlen(sp->b);
	sp->b[n] = c, sp->b[n + 1] = 0;

	/* JIS X 0212 */
	if (sp->b[0] == SS3 && strlen(sp->b) == 3) {
		if (encoding == JIS) {
			if (select_charset(sp, SELHOJO, fp) < 0)
				return -1;
			etoj(sp->b[1], sp->b[2]);
			if (rputc(sp->b[1], fp) < 0 ||
			    rputc(sp->b[2], fp) < 0)
				return -1;
		} else if (encoding == SJIS) {
			/* print TOUFU */
			sp->b[0] = TOUFU1ST;
			sp->b[1] = TOUFU2ND;
			etos(sp->b[0], sp->b[1]);
			if (rputc(sp->b[0], fp) < 0 ||
			    rputc(sp->b[1], fp) < 0)
				return -1;

		} else if (encoding == EUC) {
			if (rputs(sp->b, fp) < 0)
				return -1;
		} else if (encoding == UTF) {
			if (etou(ob, sp->b) < 0)
				strcpy(ob, "..");
			if (rputs(ob, fp) < 0)
				return -1;
		} else {
			if (eton(ob, sp->b) < 0)
				strcpy(ob, "..");
			if (rputs(ob, fp) < 0)
				return -1;
		}
		sp->b[0] = 0;
		return 2;
	}

	/* JIS X 0201R */
	if (sp->b[0] == SS2 && strlen(sp->b) == 2) {
		if (encoding == JIS) {
			if (select_charset(sp, SELKANA, fp) < 0 ||
			    rputc(sp->to_kana=='8' ? c : c&0x7f, fp) < 0)
				return -1;
		} else if (encoding == SJIS) {
			if (rputc(sp->b[1], fp) < 0)
				return -1;
		} else if (encoding == EUC) {
			if (rputs(sp->b, fp) < 0)
				return -1;
		} else if (encoding == UTF) {
			if (etou(ob, sp->b) < 0)
				strcpy(ob, ".");
			if (rputs(ob, fp) < 0)
				return -1;
		} else {
			if (eton(ob, sp->b) < 0)
				strcpy(ob, ".");
			if (rputs(ob, fp) < 0)
				return -1;
		}
		sp->b[0] = 0;
		return 1;
	}

	/* JIS X 0208 */
	if (sp->b[0] != SS2 && sp->b[0] != SS3 && strlen(sp->b) == 2) {
		if (encoding == JIS) {
			if (select_charset(sp, SELKANJI, fp) < 0)
				return -1;
			etoj(sp->b[0], sp->b[1]);
			if (rputc(sp->b[0], fp) < 0 ||
			    rputc(sp->b[1], fp) < 0)
				return -1;
		} else if (encoding == SJIS) {
			etos(sp->b[0], sp->b[1]);
			if (rputs(sp->b, fp) < 0)
				return -1;
		} else if (encoding == EUC) {
			if (rputs(sp->b, fp) < 0)
				return -1;
		} else if (encoding == UTF) {
			if (etou(ob, sp->b) < 0)
				strcpy(ob, "..");
			if (rputs(ob, fp) < 0)
				return -1;
		} else {
			if (eton(ob, sp->b) < 0)
				strcpy(ob, "..");
			if (rputs(ob, fp) < 0)
				return -1;
		}
		sp->b[0] = 0;
		return 2;
	}

	return 0;
}

int
kttputc(int c)
{
	return mbputc(&display_state, display_encoding, c, NULL);
}

/*
 * Check current JIS code condition, and output KANJI/ASCII code
 * select escape sequence to the display if nessesary.
 */
VOID
kdselectcode(next_is_k)
register int	next_is_k;		/* Is next code KANJI ? */
{
	if (display_encoding != JIS)
		return;
	(void)select_charset(&display_state, 
	    next_is_k ? SELKANJI: SELROMA, NULL);
}

/*
 * Output one byte to the file with KANJI code conversion.
 */
VOID
kputc(c, fp, kfio)
register int	c;	/* 90.07.25  Add "register". by S.Yoshida */
register FILE	*fp;
register int	kfio;
{
	(void)mbputc(&fileio_state, kfio, c, fp);
}

/*
 * Check current JIS code condition, and output KANJI/ASCII code
 * select escape sequence to the file if nessesary.
 * You must not call this function when file I/O code is not JIS.
 */
VOID
kfselectcode(fp, next_is_k)
register FILE	*fp;
register int	next_is_k;		/* Is next code KANJI ? */
{
	(void)select_charset(&fileio_state,
	    next_is_k ? SELKANJI: SELROMA, fp);
}

/*
 * Set buffer local file input code to read new file.
 */
VOID
ksetfincode(bp)
register BUFFER	*bp;
{
	if ((local_kfin = global_kexpect) == T) {
		local_kfin = NIL;
	} else if (local_kfin == NIL) {
		if (bp != NULL) {
			local_kfin = bp->b_kfio;
		}
	}
}

static int bufutoe(char *, int), bufutoe_c(const char *, int);

/*
 * Convert KANJI code form a file code (JIS/Shift-JIS/EUC) to
 * a buffer code (EUC) in the text line.
 * When file KANJI code is not decided, we check and determine it
 * to see a text line.
 */
kcodeconv(buf, len, bp)
register char	*buf;
register int	len;
register BUFFER	*bp;
{
	if (local_kfin == NIL) {	/* Not decided.	*/
		local_kfin = kcodecheck(buf, len);
	}
	if (local_kfin != NIL) {	/* Must convert. */
		if (local_kfin == UTF) {
			len = bufutoe(buf, len);
		} else if (local_kfin == JIS) {
			len = bufjtoe(buf, len);
		} else if (local_kfin == SJIS) {
			len = bufstoe(buf, len);
		}			/* EUC need not convert. */
		if (bp != NULL &&
		    bp->b_kfio == NIL) { /* set buffer local code. */
			bp->b_kfio = local_kfin;
		}
	}
	return(len);
}
/*
 * Count after converting KANJI code form a file code (JIS/Shift-JIS/EUC) to
 * a buffer code (EUC) in the text line.
 * When file KANJI code is not decided, we check and determine it
 * to see a text line.
 */
kcodecount(buf, len)
register char	*buf;
register int	len;
{
	if (local_kfin == NIL) {	/* Not decided.	*/
		local_kfin = kcodecheck(buf, len);
	}
	if (local_kfin != NIL) {	/* Must convert. */
		if (local_kfin == UTF) {
			len = bufutoe_c(buf, len);
		} else if (local_kfin == JIS) {
			len = bufjtoe_c(buf, len);
		} else if (local_kfin == SJIS) {
			len = bufstoe_c(buf, len);
		}
	}
	return(len);
}

/*
 * Check and determine what kind of KANJI code exists.
 * If no KANJI code exists, NIL (Not determine/No conversion) is
 * returned. If both EUC and Shift-JIS code is possible,
 * we think it is EUC.
 * This routine is refered to Nemacs's kanji.c.
 */
#define UTFFC(c) (((c)&0xc0)==0x80)

static int
is_valid_utf(char *s, int n)
{
	unsigned char *t, *e;

	e = (unsigned char *)s + n;
	for (t = (unsigned char *)s; t < e; ) {
		unsigned c;

		if ((t[0] & 0x80) == 0) {
			t++;
			continue;
		}
		if ((t[0] & 0xe0) == 0xc0) {
			if (e - t < 2 || !UTFFC(t[1]))
				return 0;
			c = (unsigned)(t[0] & 0x1f) << 6 | t[1] & 0x3f;
			if (c < 0x80)
				return 0;
			t += 2;
			continue;
		}
		if ((t[0] & 0xf0) == 0xe0) {
			if (e - t < 3 || !UTFFC(t[1]) || !UTFFC(t[2]))
				return 0;
			c = ((unsigned)t[0] & 0x0f) << 12 |
			    ((unsigned)t[1] & 0x3f) <<  6 | t[2] & 0x3f;
			if (c < 0x800)
				return 0;
			t += 3;
			continue;
		}
		if ((t[0] & 0xf8) == 0xf0) {
			if (e - t < 4 ||
			    !UTFFC(t[1]) || !UTFFC(t[2]) || !UTFFC(t[3]))
				return 0;
			c = ((unsigned)t[0] & 0x07) << 18 |
			    ((unsigned)t[1] & 0x3f) << 12 |
			    ((unsigned)t[2] & 0x3f) <<  6 | t[3] & 0x3f;
			if (c < 0x00010000)
				return 0;
			t += 4;
			continue;
		}
		if ((t[0] & 0xfc) == 0xf8) {
			if (e - t < 5 || !UTFFC(t[1]) ||
			    !UTFFC(t[2]) || !UTFFC(t[3]) || !UTFFC(t[4]))
				return 0;
			c = ((unsigned)t[0] & 0x03) << 24 |
			    ((unsigned)t[1] & 0x3f) << 18 |
			    ((unsigned)t[2] & 0x3f) << 12 |
			    ((unsigned)t[3] & 0x3f) <<  6 | t[4] & 0x3f;
			if (c < 0x00200000)
				return 0;
			t += 5;
			continue;
		}
		if ((t[0] & 0xfe) == 0xfc) {
			if (e - t < 6 || !UTFFC(t[1]) || !UTFFC(t[2]) ||
			    !UTFFC(t[3]) || !UTFFC(t[4]) || !UTFFC(t[5]))
				return 0;
			c = ((unsigned)t[0] & 0x03) << 30 |
			    ((unsigned)t[1] & 0x3f) << 24 |
			    ((unsigned)t[2] & 0x3f) << 18 |
			    ((unsigned)t[3] & 0x3f) << 12 |
			    ((unsigned)t[4] & 0x3f) <<  6 | t[5] & 0x3f;
			if (c < 0x04000000)
				return 0;
			t += 6;
			continue;
		}
		return 0;
	}
	return 1;
}

static int
is_possible_jis(char *s, int n)
{
	unsigned char *t, *e;

	e = (unsigned char *)s + n;
	for (t = (unsigned char *)s; t < e; t++) {
		if (t[0] == ESC &&
		    2 <= e - t && (t[1] == '$' || t[1] == '('))
			return 1;
	}
	return 0;
}

static int
is_valid_euc(const char *s, int n)
{
	unsigned char *t, *e;
	t = (unsigned char *)s, e = t + n;
	for (; t < e; t++) {
		if ((t[0] & 0x80) == 0)
			continue;

		if (t[0] == SS3) {
			int c, d;
			if (e - t < 3)
				return 0;
			c = t[1] | 0x80, d = t[2] | 0x80;
			if (!iseuc1st(c) || !iseuc2nd(c))
				return 0;
			t += 2;
		} else {
			if (e - t < 2)
				return 0;
			if (t[0] == SS2) {
				if (!iskana(t[1]))
					return 0;
			} else {
				if (!iseuc1st(t[0]) || !iseuc2nd(t[0]))
					return 0;
			}
			t ++;
		}
	}
	return 1;
}

static int
is_valid_sjis(const char *s, int n)
{
	unsigned char *t, *e;

	t = (unsigned char *)s, e = t + n;
	for (; t < e; t++) {
		if ((t[0] & 0x80) == 0)
			continue;

		if (iskana(t[0]))
			continue;
		if (!issjis1st(t[0]) || e - t < 2 ||
		    !issjis2nd(t[1]))
			return 0;
		t++;
	}
	return 1;
}

static int
kcodecheck(s, len)
char *s;
int	len;
{
	unsigned char *p = (unsigned char *)s;
	unsigned char	*endp = p + len - 1;
	register int	notjis = FALSE;

	for (; p < endp; p++) {
		if (p[0] == ESC && (p[1] == '$' || p[1] == '(') && !notjis)
			return (JIS);

		if (0x80 <= p[0]) {
		        if (!iskana(p[0]))
				notjis = TRUE;

			if (is_valid_utf(s, len))
				return UTF;
			if (is_valid_euc(s, len))
				return EUC;
			if (is_valid_sjis(s, len))
				return SJIS;
		}
	}
	return (notjis ? EUC: NIL);
}

/*
 * Convert KANJI code from JIS to EUC of the text in a buffer.
 * KANJI/ASCII selecting escape sequence is striped. So converted
 * EUC text must be shorter than sorce JIS code text.
 */
bufjtoe(j, len)
char	*j;				/* JIS code text.	*/
int	len;
{
	register int	c1, c2;		/* 90.07.25  Add "register". */
					/*		by S.Yoshida */
	register char	*e = j;		/* Converted EUC text. (same buffer) */
	register char	*endj = j + len;
	register int	kselected = SELROMA;
	char *cp, *eorg = j;

	if ( len == 0 ) return(0);
	if((cp = alloca((unsigned)(len))) == NULL) {
	    ewprintf("Could not allocate %d bytes", len);
	    return(-1);
	}
	bcopy(j, cp, len);
	j = cp;
	endj = j + len;

	while (j < endj) {
		c1 = *j++;
		if (c1 == ESC) {
			if (*j == '$' && (j[1] == '@' || j[1] == 'B')) {
			    kselected = SELKANJI;
			    j += 2;
			} else if (*j == '(' &&
				   (j[1] == 'B' || j[1] == 'J' || j[1] == 'H')) {
			  kselected = SELROMA;
			  j += 2;
			} else if (*j == '(' && j[1] == 'I') {
			    kselected = SELKANA;
			    j += 2;
			} else if (*j == '$' && j[1] == '(' && j[2] == 'D') {
			    kselected = SELHOJO;
			    j += 3;
			} else {
				*e++ = c1;
			}
		} else if (c1 == 0x0e) {
		    kselected = SELKANA;
		} else if (c1 == 0x0f) {
		    kselected = SELROMA;
		} else if (kselected == SELKANJI) {
			c2 = *j++;
			jtoe(c1, c2);
			*e++ = c1;
			*e++ = c2;
		} else if (kselected == SELHOJO) {
		    *e++ = (char)SS3;
		    c2 = *j++;
		    jtoe(c1, c2);
		    *e++ = c1;
		    *e++ = c2;
		} else if (kselected == SELKANA) {
		    *e++ = (char)SS2;
		    *e++ = (char)(c1 | 0x80);
		} else if (iskana(c1 & 0xff)) {
		    *e++ = (char)SS2;
		    *e++ = (char)c1;
		} else {
			*e++ = (char)c1;
		}
	}
	return (e - eorg);
}

/*
 * Convert KANJI code from Shift-JIS to EUC of the text in a buffer.
 */
bufstoe(p, len)
char	*p;
int	len;
{
	register int	c1, c2;		/* 90.07.25  Add "register". */
					/*              by S.Yoshida */
	register char	*porg, *tmpstr, *cp, *endcp;

	if ( len == 0 ) return(0);
	if((tmpstr = alloca((unsigned)(len))) == NULL) {
	    ewprintf("Could not allocate %d bytes", len);
	    return(-1);
	}
	cp = tmpstr;
	bcopy(p, cp, len);
	endcp = cp + len;
	porg = p;
	
	while (cp < endcp) {
		c1 = *cp++ & 0xff;
		if (iskana(c1)) {
		    *p++ = (char)SS2;
		    *p++ = (char)c1;
		} else if (issjis1st(c1)) {
			c2 = *cp++ & 0xff;
			stoe(c1, c2);
			*p++ = (char)c1;
			*p++ = (char)c2;
		} else {
			*p++ = (char)c1;
		}
	}
	return(p - porg);
}

VOID
bufetos(p, len)
char	*p;
int	len;
{
	register int	c1, c2;		/* 90.07.25  Add "register". */
					/*              by S.Yoshida */
	register char	*endp = p + len;
	register char   *q = p;

	while (p < endp) {
		c1 = *p++ & 0xff;
		if (ISKANJI(c1)) {
			if (ISHANKANA(c1))
				*q++ = *p++;
			else
				if (ISHOJO(c1)) {
					c1 = TOUFU1ST;
					c2 = TOUFU2ND;
					etos(c1,c2);
					*q++ = c1;
					*q++ = c2;
					p += 2;
				} else {
					c2 = *p++ & 0xff;
					etos(c1,c2);
					*q++ = c1;
					*q++ = c2;
				}
		} else *q++ = c1;
	}
}

static int
utlen(int c)
{
	if ((c & 0xe0) == 0xc0)
		return 2;
	if ((c & 0xf0) == 0xe0)
		return 3;
	if ((c & 0xf8) == 0xf0)
		return 4;
	if ((c & 0xfc) == 0xf8)
		return 5;
	if ((c & 0xfe) == 0xfc)
		return 6;
	return 1;
}

static int
strutoe(char *r, const char *p, int n)
{
	char *tp, *e;
	const char *endp;

	if (n == 0) {
		r != NULL && (r[0] = 0);
		return 0;
	}

	if ((tp = alloca(n)) == NULL) {
		ewprintf("Cound not allocate %d bytes", n);
		return -1;
	}
	e = tp;
	endp = p + n;
	while (p < endp) {
		char u[7];
		int nt = utlen((unsigned char)p[0]);

		memcpy(u, p, nt), u[nt] = 0;
		if (utoe(e, u) < 0)
			strcpy(e, ".");
		e += strlen(e);
		p += nt;
	}
	if (r != NULL)
		strcpy(r, tp);
	return strlen(tp);
}

static int
bufutoe(char *p, int n)
{
	return strutoe(p, p, n);
}

static int
bufutoe_c(const char *p, int n)
{
	return strutoe(NULL, p, n);
}

static int
bufjtoe_c(j, len)
char	*j;				/* JIS code text.	*/
int	len;
{
	register int	c1, c2;		/* 90.07.25  Add "register". */
					/*		by S.Yoshida */
	register int	leng;
	register char	*endj = j + len;
	register int	kselected = SELROMA;

	leng = 0;
	while (j < endj) {
		c1 = *j++;
		if (c1 == ESC) {
		    if (*j == '$' && (j[1] == '@' || j[1] == 'B')) {
			kselected = SELKANJI;
			j += 2;
		    } else if (*j == '(' &&
			       (j[1] == 'B' || j[1] == 'J' || j[1] == 'H')) {
			kselected = SELROMA;
			j += 2;
		    } else if (*j == '(' && j[1] == 'I') {
			kselected = SELKANA;
			j += 2;
		    } else if (*j == '$' && j[1] == '(' && j[2] == 'D') {
			kselected = SELHOJO;
			j += 3;
		    } else {
			leng++;
		    }
		} else if (c1 == 0x0e) {
		    kselected = SELKANA;
		} else if (c1 == 0x0f) {
		    kselected = SELROMA;
		} else if (kselected == SELKANJI) {
		    j++;
		    leng += 2;
		} else if (kselected == SELHOJO) {
		    j++;
		    leng += 3;
		} else if (kselected == SELKANA) {
		    leng += 2;
		} else if (iskana(c1 & 0xff)) {
		    leng += 2;
		} else {
		    leng++;
		}
	}
	return (leng);
}

int
bufstoe_c(p, len)
char	*p;
int	len;
{
	register int	c1;		/* 90.07.25  Add "register". */
					/*              by S.Yoshida */
	register char	*endp;

	endp = p + len;

	len=0;
	while (p < endp) {
		c1 = *p++ & 0xff;
		if (iskana(c1)) {
			len += 2;
		} else {
			if (issjis1st(c1)) {
				p++;
				len += 2;
			} else {
				len++;
			}
		}
	}
	return(len);
}

/*
 * Is current position char KANJI ?
 */
iskanji()
{
	return(ISKANJI(lgetc(curwp->w_dotp, curwp->w_doto)));
}

/*
 * Is current position char Hojo KANJI ?
 */
ishojo()
{
	return(ISHOJO(lgetc(curwp->w_dotp, curwp->w_doto)));
}

/*
 * Is this KANJI word char ?
 * This routine return TRUE when c1 is ASCII char,
 * but it doesn't happen now.
 */
iskword(c1, c2)
register int	c1;
register int	c2;
{
	return(charcategory(c1, c2) != CAT_NONE);
}

/*
 * Get char category of current position char.
 */
getcategory()
{
	return(charcategory(lgetc(curwp->w_dotp, curwp->w_doto),
			    lgetc(curwp->w_dotp, curwp->w_doto + 1)));
}

int	cur_cat;			/* Current char category.	*/
int	cur_dir;			/* Current search direction.	*/

/*
 * Set char category of search start position char,
 * and set search direction.
 */
VOID
initcategory(dir)
int	dir;
{
	cur_cat = getcategory();
	cur_dir = dir;
}

/*
 * Is it in a same category ?
 * return TRUE if category of current position char is same as
 * start position char's one (now in a word), or it is HIRAGANA
 * added to the end of that word.
 */
incategory()
{
	register int	cat;

	if (curwp->w_doto == llength(curwp->w_dotp)) {
		return FALSE;
	}
	cat = getcategory();
	if (cur_dir == 1) {		/* Direction is forward. */
		if (cat == CAT_HIRAGANA) { /* Now start added HIRAGANA part. */
			cur_cat = cat;
		}
	} else {			/* Direcrion is backward. */
		if (cur_cat == CAT_HIRAGANA && cat != cur_cat) {
			/* Now end added HIRAGANA part and start word part. */
			cur_cat = cat;
		}
	}
	return(cat == cur_cat);
}

/*
 * Return char category of one char.
 */
charcategory(c1, c2)
register int	c1;
register int	c2;
{
	switch (c1) {
	    case 0xa1:
		if (c2 >= 0xb8 && c2 <= 0xbb) {
			return(CAT_KANJI);
		} else if (c2 >= 0xb3 && c2 <= 0xb6) {
			return(CAT_HIRAGANA);
		} else if (c2 == 0xbc) {
			return(CAT_KATAKANA);
		}
		/* else it is KIGOU. */
	    case 0xa2:
	    case 0xa8:
		return(CAT_KIGOU);
		/* Not reached */
	    case 0xa3:
		return(CAT_EISUUJI);
		/* Not reached */
	    case 0xa4:
		return(CAT_HIRAGANA);
		/* Not reached */
	    case 0xa5:
		return(CAT_KATAKANA);
		/* Not reached */
	    case 0xa6:
		return(CAT_GREEK);
		/* Not reached */
	    case 0xa7:
		return(CAT_RUSSIAN);
		/* Not reached */
	    default:
		if (c1 <= 0x7f) {
			return(CAT_ASCII);
		} else if (c1 >= 0xb0) {
			return(CAT_KANJI);
		}
		return(CAT_NONE);
		/* Not reached */
	}
}
